from typing import Iterator

import dlt
import pandas as pd  # type: ignore[import-untyped]
import pendulum
import requests
from dlt.sources import DltResource
from dlt.sources.helpers.requests import Client
from pendulum.date import Date


@dlt.source(max_table_nesting=0)
def applovin_max_source(
    start_date: str,
    application: str,
    api_key: str,
    end_date: str | None,
) -> DltResource:
    @dlt.resource(
        name="ad_revenue",
        write_disposition="merge",
        merge_key="_partition_date",
    )
    def fetch_ad_revenue_report(
        dateTime=(
            dlt.sources.incremental(
                "_partition_date",
                initial_value=start_date,
                end_value=end_date,
                range_start="closed",
                range_end="closed",
            )
        ),
    ) -> Iterator[dict]:
        url = "https://r.applovin.com/max/userAdRevenueReport"
        start_date = pendulum.from_format(dateTime.last_value, "YYYY-MM-DD").date()
        if dateTime.end_value is None:
            end_date = (pendulum.yesterday("UTC")).date()
        else:
            end_date = pendulum.from_format(dateTime.end_value, "YYYY-MM-DD").date()
        yield get_data(
            url=url,
            start_date=start_date,
            end_date=end_date,
            application=application,
            api_key=api_key,
        )

    return fetch_ad_revenue_report


def create_client() -> requests.Session:
    return Client(
        request_timeout=10.0,
        raise_for_status=False,
        retry_condition=retry_on_limit,
        request_max_attempts=12,
    ).session


def retry_on_limit(
    response: requests.Response | None, exception: BaseException | None
) -> bool:
    if response is None:
        return False
    return response.status_code == 429


def get_data(
    url: str, start_date: Date, end_date: Date, application: str, api_key: str
):
    client = create_client()
    platforms = ["ios", "android", "fireos"]
    current_date = start_date
    while current_date <= end_date:
        for platform in platforms:
            params = {
                "api_key": api_key,
                "date": current_date.strftime("%Y-%m-%d"),
                "platform": platform,
                "application": application,
                "aggregated": "false",
            }

            response = client.get(url=url, params=params)

            if response.status_code == 400:
                raise ValueError(response.text)

            if response.status_code != 200:
                continue

            response_url = response.json().get("ad_revenue_report_url")
            df = pd.read_csv(response_url)
            df["Date"] = pd.to_datetime(df["Date"])
            df["_partition_date"] = df["Date"].dt.strftime("%Y-%m-%d")
            yield df

        current_date = current_date.add(days=1)



"""This source uses Stripe API and dlt to load data such as Customer, Subscription, Event etc. to the database and to calculate the MRR and churn rate."""

from typing import Any, Dict, Generator, Iterable, Optional, Tuple

import dlt
import stripe
from dlt.sources import DltResource
from pendulum import DateTime

from .helpers import pagination, transform_date
from .settings import ENDPOINTS, INCREMENTAL_ENDPOINTS


@dlt.source(max_table_nesting=0)
def stripe_source(
    endpoints: Tuple[str, ...] = ENDPOINTS,
    stripe_secret_key: str = dlt.secrets.value,
    start_date: Optional[DateTime] = None,
    end_date: Optional[DateTime] = None,
) -> Iterable[DltResource]:
    """
    Retrieves data from the Stripe API for the specified endpoints.

    For all endpoints, Stripe API responses do not provide the key "updated",
    so in most cases, we are forced to load the data in 'replace' mode.
    This source is suitable for all types of endpoints, including 'Events', 'Invoice', etc.
    but these endpoints can also be loaded in incremental mode (see source incremental_stripe_source).

    Args:
        endpoints (Tuple[str, ...]): A tuple of endpoint names to retrieve data from. Defaults to most popular Stripe API endpoints.
        stripe_secret_key (str): The API access token for authentication. Defaults to the value in the `dlt.secrets` object.
        start_date (Optional[DateTime]): An optional start date to limit the data retrieved. Format: datetime(YYYY, MM, DD). Defaults to None.
        end_date (Optional[DateTime]): An optional end date to limit the data retrieved. Format: datetime(YYYY, MM, DD). Defaults to None.

    Returns:
        Iterable[DltResource]: Resources with data that was created during the period greater than or equal to 'start_date' and less than 'end_date'.
    """
    stripe.api_key = stripe_secret_key
    stripe.api_version = "2022-11-15"

    def stripe_resource(
        endpoint: str,
    ) -> Generator[Dict[Any, Any], Any, None]:
        yield from pagination(endpoint, start_date, end_date)

    for endpoint in endpoints:
        yield dlt.resource(
            stripe_resource,
            name=endpoint,
            write_disposition="replace",
        )(endpoint)


@dlt.source
def incremental_stripe_source(
    endpoints: Tuple[str, ...] = INCREMENTAL_ENDPOINTS,
    stripe_secret_key: str = dlt.secrets.value,
    initial_start_date: Optional[DateTime] = None,
    end_date: Optional[DateTime] = None,
) -> Iterable[DltResource]:
    """
    As Stripe API does not include the "updated" key in its responses,
    we are only able to perform incremental downloads from endpoints where all objects are uneditable.
    This source yields the resources with incremental loading based on "append" mode.
    You will load only the newest data without duplicating and without downloading a huge amount of data each time.

    Args:
        endpoints (tuple): A tuple of endpoint names to retrieve data from. Defaults to Stripe API endpoints with uneditable data.
        stripe_secret_key (str): The API access token for authentication. Defaults to the value in the `dlt.secrets` object.
        initial_start_date (Optional[DateTime]): An optional parameter that specifies the initial value for dlt.sources.incremental.
                            If parameter is not None, then load only data that were created after initial_start_date on the first run.
                            Defaults to None. Format: datetime(YYYY, MM, DD).
        end_date (Optional[DateTime]): An optional end date to limit the data retrieved.
                  Defaults to None. Format: datetime(YYYY, MM, DD).
    Returns:
        Iterable[DltResource]: Resources with only that data has not yet been loaded.
    """
    stripe.api_key = stripe_secret_key
    stripe.api_version = "2022-11-15"
    start_date_unix = (
        transform_date(initial_start_date) if initial_start_date is not None else -1
    )

    def incremental_resource(
        endpoint: str,
        created: Optional[Any] = dlt.sources.incremental(
            "created",
            initial_value=start_date_unix,
            range_end="closed",
            range_start="closed",
        ),
    ) -> Generator[Dict[Any, Any], Any, None]:
        start_value = created.last_value
        yield from pagination(endpoint, start_date=start_value, end_date=end_date)

    for endpoint in endpoints:
        yield dlt.resource(
            incremental_resource,
            name=endpoint,
            write_disposition="append",
            primary_key="id",
        )(endpoint)
